
from spinnman.messages.eieio.data_messages.eieio_16bit\
    .eieio_16bit_data_message import EIEIO16BitDataMessage
from spinnman.messages.eieio.data_messages.eieio_16bit\
    .eieio_16bit_lower_key_prefix_data_message\
    import EIEIO16BitLowerKeyPrefixDataMessage
from spinnman.messages.eieio.data_messages.eieio_16bit\
    .eieio_16bit_upper_key_prefix_data_message\
    import EIEIO16BitUpperKeyPrefixDataMessage
from spinnman.messages.eieio.data_messages.eieio_16bit\
    .eieio_16bit_payload_prefix_data_message\
    import EIEIO16BitPayloadPrefixDataMessage
from spinnman.messages.eieio.data_messages.eieio_16bit\
    .eieio_16bit_payload_prefix_lower_key_prefix_data_message\
    import EIEIO16BitPayloadPrefixLowerKeyPrefixDataMessage
from spinnman.messages.eieio.data_messages.eieio_16bit\
    .eieio_16bit_payload_prefix_upper_key_prefix_data_message\
    import EIEIO16BitPayloadPrefixUpperKeyPrefixDataMessage
from spinnman.messages.eieio.data_messages.eieio_16bit\
    .eieio_16bit_timed_payload_prefix_data_message\
    import EIEIO16BitTimedPayloadPrefixDataMessage
from spinnman.messages.eieio.data_messages.eieio_16bit\
    .eieio_16bit_timed_payload_prefix_lower_key_prefix_data_message\
    import EIEIO16BitTimedPayloadPrefixLowerKeyPrefixDataMessage
from spinnman.messages.eieio.data_messages.eieio_16bit\
    .eieio_16bit_timed_payload_prefix_upper_key_prefix_data_message\
    import EIEIO16BitTimedPayloadPrefixUpperKeyPrefixDataMessage
from spinnman.messages.eieio.data_messages.eieio_16bit_with_payload\
    .eieio_16bit_with_payload_data_message\
    import EIEIO16BitWithPayloadDataMessage
from spinnman.messages.eieio.data_messages.eieio_16bit_with_payload\
    .eieio_16bit_with_payload_lower_key_prefix_data_message\
    import EIEIO16BitWithPayloadLowerKeyPrefixDataMessage
from spinnman.messages.eieio.data_messages.eieio_16bit_with_payload\
    .eieio_16bit_with_payload_upper_key_prefix_data_message\
    import EIEIO16BitWithPayloadUpperKeyPrefixDataMessage
from spinnman.messages.eieio.data_messages.eieio_16bit_with_payload\
    .eieio_16bit_with_payload_payload_prefix_data_message\
    import EIEIO16BitWithPayloadPayloadPrefixDataMessage
from spinnman.messages.eieio.data_messages.eieio_16bit_with_payload\
    .eieio_16bit_with_payload_payload_prefix_lower_key_prefix_data_message\
    import EIEIO16BitWithPayloadPayloadPrefixLowerKeyPrefixDataMessage
from spinnman.messages.eieio.data_messages.eieio_16bit_with_payload\
    .eieio_16bit_with_payload_payload_prefix_upper_key_prefix_data_message\
    import EIEIO16BitWithPayloadPayloadPrefixUpperKeyPrefixDataMessage
from spinnman.messages.eieio.data_messages.eieio_16bit_with_payload\
    .eieio_16bit_with_payload_timed_data_message\
    import EIEIO16BitWithPayloadTimedDataMessage
from spinnman.messages.eieio.data_messages.eieio_16bit_with_payload\
    .eieio_16bit_with_payload_timed_lower_key_prefix_data_message\
    import EIEIO16BitWithPayloadTimedLowerKeyPrefixDataMessage
from spinnman.messages.eieio.data_messages.eieio_16bit_with_payload\
    .eieio_16bit_with_payload_timed_upper_key_prefix_data_message\
    import EIEIO16BitWithPayloadTimedUpperKeyPrefixDataMessage
from spinnman.messages.eieio.data_messages.eieio_32bit\
    .eieio_32bit_data_message import EIEIO32BitDataMessage
from spinnman.messages.eieio.data_messages.eieio_32bit\
    .eieio_32bit_lower_key_prefix_data_message\
    import EIEIO32BitLowerKeyPrefixDataMessage
from spinnman.messages.eieio.data_messages.eieio_32bit\
    .eieio_32bit_upper_key_prefix_data_message\
    import EIEIO32BitUpperKeyPrefixDataMessage
from spinnman.messages.eieio.data_messages.eieio_32bit\
    .eieio_32bit_payload_prefix_data_message\
    import EIEIO32BitPayloadPrefixDataMessage
from spinnman.messages.eieio.data_messages.eieio_32bit\
    .eieio_32bit_payload_prefix_lower_key_prefix_data_message\
    import EIEIO32BitPayloadPrefixLowerKeyPrefixDataMessage
from spinnman.messages.eieio.data_messages.eieio_32bit\
    .eieio_32bit_payload_prefix_upper_key_prefix_data_message\
    import EIEIO32BitPayloadPrefixUpperKeyPrefixDataMessage
from spinnman.messages.eieio.data_messages.eieio_32bit\
    .eieio_32bit_timed_payload_prefix_data_message\
    import EIEIO32BitTimedPayloadPrefixDataMessage
from spinnman.messages.eieio.data_messages.eieio_32bit\
    .eieio_32bit_timed_payload_prefix_lower_key_prefix_data_message\
    import EIEIO32BitTimedPayloadPrefixLowerKeyPrefixDataMessage
from spinnman.messages.eieio.data_messages.eieio_32bit\
    .eieio_32bit_timed_payload_prefix_upper_key_prefix_data_message\
    import EIEIO32BitTimedPayloadPrefixUpperKeyPrefixDataMessage
from spinnman.messages.eieio.data_messages.eieio_32bit_with_payload\
    .eieio_32bit_with_payload_data_message\
    import EIEIO32BitWithPayloadDataMessage
from spinnman.messages.eieio.data_messages.eieio_32bit_with_payload\
    .eieio_32bit_with_payload_lower_key_prefix_data_message\
    import EIEIO32BitWithPayloadLowerKeyPrefixDataMessage
from spinnman.messages.eieio.data_messages.eieio_32bit_with_payload\
    .eieio_32bit_with_payload_upper_key_prefix_data_message\
    import EIEIO32BitWithPayloadUpperKeyPrefixDataMessage
from spinnman.messages.eieio.data_messages.eieio_32bit_with_payload\
    .eieio_32bit_with_payload_payload_prefix_data_message\
    import EIEIO32BitWithPayloadPayloadPrefixDataMessage
from spinnman.messages.eieio.data_messages.eieio_32bit_with_payload\
    .eieio_32bit_with_payload_payload_prefix_lower_key_prefix_data_message\
    import EIEIO32BitWithPayloadPayloadPrefixLowerKeyPrefixDataMessage
from spinnman.messages.eieio.data_messages.eieio_32bit_with_payload\
    .eieio_32bit_with_payload_payload_prefix_upper_key_prefix_data_message\
    import EIEIO32BitWithPayloadPayloadPrefixUpperKeyPrefixDataMessage
from spinnman.messages.eieio.data_messages.eieio_32bit_with_payload\
    .eieio_32bit_with_payload_timed_data_message\
    import EIEIO32BitWithPayloadTimedDataMessage
from spinnman.messages.eieio.data_messages.eieio_32bit_with_payload\
    .eieio_32bit_with_payload_timed_lower_key_prefix_data_message\
    import EIEIO32BitWithPayloadTimedLowerKeyPrefixDataMessage
from spinnman.messages.eieio.data_messages.eieio_32bit_with_payload\
    .eieio_32bit_with_payload_timed_upper_key_prefix_data_message\
    import EIEIO32BitWithPayloadTimedUpperKeyPrefixDataMessage
from spinnman.messages.eieio.eieio_prefix import EIEIOPrefix
from spinnman.messages.eieio.data_messages.eieio_data_message\
    import EIEIODataMessage
from spinnman.messages.eieio.data_messages.eieio_data_header\
    import EIEIODataHeader
from spinnman.messages.eieio.eieio_type import EIEIOType
from spinnman.messages.eieio.data_messages.eieio_without_payload_data_message\
    import EIEIOWithoutPayloadDataMessage
from spinnman.messages.eieio.data_messages.eieio_with_payload_data_message\
    import EIEIOWithPayloadDataMessage


def _read_16_bit_message(prefix, payload_base, prefix_type, is_time,
                         data, offset, eieio_header):
    """ Return a packet containing 16 bit elements
    """
    if payload_base is None:
        if prefix is None:
            return EIEIO16BitDataMessage(eieio_header.count, data, offset)
        elif prefix_type == EIEIOPrefix.LOWER_HALF_WORD:
            return EIEIO16BitLowerKeyPrefixDataMessage(
                prefix, eieio_header.count, data, offset)
        elif prefix_type == EIEIOPrefix.UPPER_HALF_WORD:
            return EIEIO16BitUpperKeyPrefixDataMessage(
                prefix, eieio_header.count, data, offset)
    elif payload_base is not None and not is_time:
        if prefix is None:
            return EIEIO16BitPayloadPrefixDataMessage(
                payload_base, eieio_header.count, data, offset)
        elif prefix_type == EIEIOPrefix.LOWER_HALF_WORD:
            return EIEIO16BitPayloadPrefixLowerKeyPrefixDataMessage(
                prefix, payload_base, eieio_header.count, data, offset)
        elif prefix_type == EIEIOPrefix.UPPER_HALF_WORD:
            return EIEIO16BitPayloadPrefixUpperKeyPrefixDataMessage(
                prefix, payload_base, eieio_header.count, data, offset)
    elif payload_base is not None and is_time:
        if prefix is None:
            return EIEIO16BitTimedPayloadPrefixDataMessage(
                payload_base, eieio_header.count, data, offset)
        elif prefix_type == EIEIOPrefix.LOWER_HALF_WORD:
            return EIEIO16BitTimedPayloadPrefixLowerKeyPrefixDataMessage(
                prefix, payload_base, eieio_header.count, data, offset)
        elif prefix_type == EIEIOPrefix.UPPER_HALF_WORD:
            return EIEIO16BitTimedPayloadPrefixUpperKeyPrefixDataMessage(
                prefix, payload_base, eieio_header.count, data, offset)
    return EIEIOWithoutPayloadDataMessage(eieio_header, data, offset)


def _read_16_bit_payload_message(prefix, payload_base, prefix_type,
                                 is_time, data, offset, eieio_header):
    """ Return a packet containing 16 bit elements and payload
    """
    if payload_base is None and not is_time:
        if prefix is None:
            return EIEIO16BitWithPayloadDataMessage(
                eieio_header.count, data, offset)
        elif prefix_type == EIEIOPrefix.LOWER_HALF_WORD:
            return EIEIO16BitWithPayloadLowerKeyPrefixDataMessage(
                prefix, eieio_header.count, data, offset)
        elif prefix_type == EIEIOPrefix.UPPER_HALF_WORD:
            return EIEIO16BitWithPayloadUpperKeyPrefixDataMessage(
                prefix, eieio_header.count, data, offset)
    elif payload_base is not None and not is_time:
        if prefix is None:
            return EIEIO16BitWithPayloadPayloadPrefixDataMessage(
                payload_base, eieio_header.count, data, offset)
        elif prefix_type == EIEIOPrefix.LOWER_HALF_WORD:
            return EIEIO16BitWithPayloadPayloadPrefixLowerKeyPrefixDataMessage(
                prefix, payload_base, eieio_header.count, data, offset)
        elif prefix_type == EIEIOPrefix.UPPER_HALF_WORD:
            return EIEIO16BitWithPayloadPayloadPrefixUpperKeyPrefixDataMessage(
                prefix, payload_base, eieio_header.count, data, offset)
    elif payload_base is None and is_time:
        if prefix is None:
            return EIEIO16BitWithPayloadTimedDataMessage(
                eieio_header.count, data, offset)
        elif prefix_type == EIEIOPrefix.LOWER_HALF_WORD:
            return EIEIO16BitWithPayloadTimedLowerKeyPrefixDataMessage(
                prefix, eieio_header.count, data, offset)
        elif prefix_type == EIEIOPrefix.UPPER_HALF_WORD:
            return EIEIO16BitWithPayloadTimedUpperKeyPrefixDataMessage(
                prefix, eieio_header.count, data, offset)
    return EIEIOWithPayloadDataMessage(eieio_header, data, offset)


def _read_32_bit_message(prefix, payload_base, prefix_type, is_time,
                         data, offset, eieio_header):
    """ Return a packet containing 32 bit elements
    """
    if payload_base is None:
        if prefix is None:
            return EIEIO32BitDataMessage(eieio_header.count, data, offset)
        elif prefix_type == EIEIOPrefix.LOWER_HALF_WORD:
            return EIEIO32BitLowerKeyPrefixDataMessage(
                prefix, eieio_header.count, data, offset)
        elif prefix_type == EIEIOPrefix.UPPER_HALF_WORD:
            return EIEIO32BitUpperKeyPrefixDataMessage(
                prefix, eieio_header.count, data, offset)
    elif payload_base is not None and not is_time:
        if prefix is None:
            return EIEIO32BitPayloadPrefixDataMessage(
                payload_base, eieio_header.count, data, offset)
        elif prefix_type == EIEIOPrefix.LOWER_HALF_WORD:
            return EIEIO32BitPayloadPrefixLowerKeyPrefixDataMessage(
                prefix, payload_base, eieio_header.count, data, offset)
        elif prefix_type == EIEIOPrefix.UPPER_HALF_WORD:
            return EIEIO32BitPayloadPrefixUpperKeyPrefixDataMessage(
                prefix, payload_base, eieio_header.count, data, offset)
    elif payload_base is not None and is_time:
        if prefix is None:
            return EIEIO32BitTimedPayloadPrefixDataMessage(
                payload_base, eieio_header.count, data, offset)
        elif prefix_type == EIEIOPrefix.LOWER_HALF_WORD:
            return EIEIO32BitTimedPayloadPrefixLowerKeyPrefixDataMessage(
                prefix, payload_base, eieio_header.count, data, offset)
        elif prefix_type == EIEIOPrefix.UPPER_HALF_WORD:
            return EIEIO32BitTimedPayloadPrefixUpperKeyPrefixDataMessage(
                prefix, payload_base, eieio_header.count, data, offset)
    return EIEIOWithoutPayloadDataMessage(eieio_header, data, offset)


def _read_32_bit_payload_message(prefix, payload_base, prefix_type,
                                 is_time, data, offset, eieio_header):
    """ Return a packet containing 32 bit elements and payload
    """
    if payload_base is None and not is_time:
        if prefix is None:
            return EIEIO32BitWithPayloadDataMessage(
                eieio_header.count, data, offset)
        elif prefix_type == EIEIOPrefix.LOWER_HALF_WORD:
            return EIEIO32BitWithPayloadLowerKeyPrefixDataMessage(
                prefix, eieio_header.count, data, offset)
        elif prefix_type == EIEIOPrefix.UPPER_HALF_WORD:
            return EIEIO32BitWithPayloadUpperKeyPrefixDataMessage(
                prefix, eieio_header.count, data, offset)
    elif payload_base is not None and not is_time:
        if prefix is None:
            return EIEIO32BitWithPayloadPayloadPrefixDataMessage(
                payload_base, eieio_header.count, data, offset)
        elif prefix_type == EIEIOPrefix.LOWER_HALF_WORD:
            return EIEIO32BitWithPayloadPayloadPrefixLowerKeyPrefixDataMessage(
                prefix, payload_base, eieio_header.count, data, offset)
        elif prefix_type == EIEIOPrefix.UPPER_HALF_WORD:
            return EIEIO32BitWithPayloadPayloadPrefixUpperKeyPrefixDataMessage(
                prefix, payload_base, eieio_header.count, data, offset)
    elif payload_base is None and is_time:
        if prefix is None:
            return EIEIO32BitWithPayloadTimedDataMessage(
                eieio_header.count, data, offset)
        elif prefix_type == EIEIOPrefix.LOWER_HALF_WORD:
            return EIEIO32BitWithPayloadTimedLowerKeyPrefixDataMessage(
                prefix, eieio_header.count, data, offset)
        elif prefix_type == EIEIOPrefix.UPPER_HALF_WORD:
            return EIEIO32BitWithPayloadTimedUpperKeyPrefixDataMessage(
                prefix, eieio_header.count, data, offset)
    return EIEIOWithPayloadDataMessage(eieio_header, data, offset)


def read_eieio_data_message(data, offset):
    """ Reads the content of an EIEIO data message and returns an object\
        identifying the data which was contained in the packet

    :param data: data received from the network as a bytestring
    :type data: str
    :param offset: offset at which the parsing operation should start
    :type offset: int
    :return: an object which inherits from EIEIODataMessage which contains\
            parsed data received from the network
    :rtype:\
            :py:class:`spinnman.messages.eieio.data_messages.eieio_data_message.EIEIODataMessage`
    """
    eieio_header = EIEIODataHeader.from_bytestring(data, offset)
    offset += eieio_header.size
    eieio_type = eieio_header.eieio_type
    prefix = eieio_header.prefix
    payload_base = eieio_header.payload_base
    prefix_type = eieio_header.prefix_type
    is_time = eieio_header.is_time
    if eieio_type == EIEIOType.KEY_16_BIT:
        return _read_16_bit_message(
            prefix, payload_base, prefix_type, is_time, data, offset,
            eieio_header)
    elif eieio_type == EIEIOType.KEY_PAYLOAD_16_BIT:
        return _read_16_bit_payload_message(
            prefix, payload_base, prefix_type, is_time, data, offset,
            eieio_header)
    elif eieio_type == EIEIOType.KEY_32_BIT:
        return _read_32_bit_message(
            prefix, payload_base, prefix_type, is_time, data, offset,
            eieio_header)
    elif eieio_type == EIEIOType.KEY_PAYLOAD_32_BIT:
        return _read_32_bit_payload_message(
            prefix, payload_base, prefix_type, is_time, data, offset,
            eieio_header)
    return EIEIODataMessage(eieio_header, data, offset)



from enum import Enum


class SCPLEDAction(Enum):
    """ The SCP LED actions
    """

    TOGGLE = (1, "Toggle the LED status")
    OFF = (2, "Turn the LED off")
    ON = (3, "Turn the LED on")

    def __new__(cls, value, doc=""):
        obj = object.__new__(cls)
        obj._value_ = value
        return obj

    def __init__(self, value, doc=""):
        self._value_ = value
        self.__doc__ = doc



from spinnman.messages.scp.abstract_messages.abstract_scp_response \
    import AbstractSCPResponse
from spinnman.messages.scp.enums.scp_result import SCPResult
from spinnman.exceptions import SpinnmanUnexpectedResponseCodeException


class SCPCheckOKResponse(AbstractSCPResponse):
    """ An SCP response to a request which returns nothing other than OK
    """

    def __init__(self, operation, command):
        """

        :param operation: The operation being performed
        :type operation: str
        :param command: The command that was sent
        :type command: str
        """
        super(SCPCheckOKResponse, self).__init__()
        self._operation = operation
        self._command = command

    def read_data_bytestring(self, data, offset):
        """ See\
            :py:meth:`spinnman.messages.scp.abstract_scp_response.AbstractSCPResponse.read_data_bytestring`
        """
        result = self.scp_response_header.result
        if result != SCPResult.RC_OK:
            raise SpinnmanUnexpectedResponseCodeException(
                self._operation, self._command, result.name)



from enum import Enum


class EIEIOType(Enum):
    """ Possible types of EIEIO packets
    """
    KEY_16_BIT = (0, 2, 0, "Indicates that data is keys which are 16 bits")
    KEY_PAYLOAD_16_BIT = (
        1, 2, 2, "Indicates that data is keys and payloads of 16 bits")
    KEY_32_BIT = (2, 4, 0, "Indicates that data is keys of 32 bits")
    KEY_PAYLOAD_32_BIT = (
        3, 4, 4, "Indicates that data is keys and payloads of 32 bits")

    def __new__(cls, value, key_bytes, payload_bytes, doc=""):
        obj = object.__new__(cls)
        obj._value_ = value
        return obj

    def __init__(self, value, key_bytes, payload_bytes, doc=""):
        self._value_ = value
        self._key_bytes = key_bytes
        self._payload_bytes = payload_bytes
        self.__doc__ = doc

    @property
    def key_bytes(self):
        """ The number of bytes used by each key element

        :rtype: int
        """
        return self._key_bytes

    @property
    def payload_bytes(self):
        """ The number of bytes used by each payload element

        :rtype: int
        """
        return self._payload_bytes

    @property
    def max_value(self):
        """ The maximum value of the key or payload (if there is a payload)

        :rtype: int
        """
        return (1 << (self._key_bytes * 8)) - 1



from spinnman.exceptions import SpinnmanInvalidParameterException
from spinnman.messages.eieio.eieio_type import EIEIOType
from spinnman.messages.eieio.data_messages.abstract_eieio_data_element \
    import AbstractEIEIODataElement

import struct


class EIEIOKeyDataElement(AbstractEIEIODataElement):
    """ A data element that contains just a key
    """

    def __init__(self, key):
        self._key = key

    @property
    def key(self):
        return self._key

    def get_bytestring(self, eieio_type):
        if eieio_type.payload_bytes != 0:
            raise SpinnmanInvalidParameterException(
                "eieio_type", eieio_type,
                "The type specifies a payload, but this element has no"
                " payload")
        if eieio_type == EIEIOType.KEY_16_BIT:
            return struct.pack("<H", self._key)
        elif eieio_type == EIEIOType.KEY_32_BIT:
            return struct.pack("<I", self._key)
        else:
            raise SpinnmanInvalidParameterException(
                "eieio_type", eieio_type, "Unknown type")

    def __str__(self):
        return "EIEIOKeyDataElement:{}".format(hex(self._key))

    def __repr__(self):
        return self.__str__()



from spinnman.messages.scp.abstract_messages.abstract_scp_request \
    import AbstractSCPRequest
from spinnman.messages.sdp.sdp_header import SDPHeader
from spinnman.messages.sdp.sdp_flag import SDPFlag
from spinnman.messages.scp.scp_request_header import SCPRequestHeader
from spinnman.messages.scp.enums.scp_command import SCPCommand
from spinnman.exceptions import SpinnmanInvalidParameterException
from spinnman.messages.scp.impl.scp_check_ok_response import SCPCheckOKResponse

_ALL_CORE_MASK = 0xFFFF
_APP_MASK = 0xFF


def _get_data(app_id, signal):
    data = (_APP_MASK << 8) | app_id
    data += signal.value << 16
    return data


class SCPSendSignalRequest(AbstractSCPRequest):
    """ An SCP Request to send a signal to cores
    """

    def __init__(self, app_id, signal):
        """

        :param app_id: The id of the application, between 0 and 255
        :type app_id: int
        :param signal: The signal to send
        :type signal: :py:class:`spinnman.messages.scp.scp_signal.SCPSignal`
        :raise spinnman.exceptions.SpinnmanInvalidParameterException: If\
            app_id is out of range
        """

        if app_id < 0 or app_id > 255:
            raise SpinnmanInvalidParameterException(
                "app_id", str(app_id), "Must be between 0 and 255")

        super(SCPSendSignalRequest, self).__init__(
            SDPHeader(
                flags=SDPFlag.REPLY_EXPECTED, destination_port=0,
                destination_cpu=0,
                destination_chip_x=self.DEFAULT_DEST_X_COORD,
                destination_chip_y=self.DEFAULT_DEST_Y_COORD),
            SCPRequestHeader(command=SCPCommand.CMD_SIG),
            argument_1=signal.signal_type.value,
            argument_2=_get_data(app_id, signal),
            argument_3=_ALL_CORE_MASK)

    def get_scp_response(self):
        return SCPCheckOKResponse("Send Signal", "CMD_SIG")



from spinnman.messages.scp.abstract_messages.abstract_scp_request\
    import AbstractSCPRequest
from spinnman.messages.sdp.sdp_header import SDPHeader
from spinnman.messages.sdp.sdp_flag import SDPFlag
from spinnman.messages.scp.scp_request_header import SCPRequestHeader
from spinnman.messages.scp.enums.scp_command import SCPCommand
from spinnman.messages.scp.impl.scp_iptag_info_response\
    import SCPIPTagInfoResponse

_IPTAG_INFO = 4
_IPTAG_MAX = 255


class SCPTagInfoRequest(AbstractSCPRequest):
    """ An SCP Request information about IP tags
    """

    def __init__(self, x, y):
        """
        :param x: The x-coordinate of a chip, between 0 and 255
        :type x: int
        :param y: The y-coordinate of a chip, between 0 and 255
        :type y: int
        """
        super(SCPTagInfoRequest, self).__init__(
            SDPHeader(
                flags=SDPFlag.REPLY_EXPECTED, destination_port=0,
                destination_cpu=0, destination_chip_x=x,
                destination_chip_y=y),
            SCPRequestHeader(command=SCPCommand.CMD_IPTAG),
            argument_1=(_IPTAG_INFO << 16),
            argument_2=_IPTAG_MAX)

    def get_scp_response(self):
        return SCPIPTagInfoResponse()



from six import add_metaclass

from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spinnman.connections.abstract_classes.abstract_connection \
    import AbstractConnection


@add_metaclass(AbstractBase)
class AbstractSpinnakerBootSender(AbstractConnection):
    """ A sender of Spinnaker Boot messages
    """

    __slots__ = ()

    @abstractmethod
    def send_boot_message(self, boot_message):
        """ Sends a SpiNNaker boot message using this connection

        :param boot_message: The message to be sent
        :type boot_message:\
                    :py:class:`spinnman.messages.spinnaker_boot.spinnaker_boot_message.SpinnakerBootMessage`
        :return: Nothing is returned
        :rtype: None
        :raise spinnman.exceptions.SpinnmanIOException: If there is an error\
                    sending the message
        """
        pass



from spinnman.messages.scp.abstract_messages.abstract_scp_request \
    import AbstractSCPRequest
from spinnman.messages.sdp.sdp_header import SDPHeader
from spinnman.messages.sdp.sdp_flag import SDPFlag
from spinnman.messages.scp.scp_request_header import SCPRequestHeader
from spinnman.messages.scp.enums.scp_command import SCPCommand
from spinnman.messages.scp.impl.scp_count_state_response import \
    SCPCountStateResponse

_ALL_CORE_MASK = 0xFFFF
_COUNT_OPERATION = 1
_COUNT_MODE = 2
_COUNT_SIGNAL_TYPE = 1
_APP_MASK = 0xFF


def _get_data(app_id, state):
    data = (_APP_MASK << 8) | app_id
    data += (_COUNT_OPERATION << 22) | (_COUNT_MODE << 20)
    data += state.value << 16
    return data


class SCPCountStateRequest(AbstractSCPRequest):
    """ An SCP Request to get a count of the cores in a particular state
    """

    def __init__(self, app_id, state):
        """

        :param app_id: The id of the application, between 0 and 255
        :type app_id: int
        :param state: The state to count
        :type state: :py:class:`spinnman.model.cpu_state.CPUState`
        """
        super(SCPCountStateRequest, self).__init__(
            SDPHeader(
                flags=SDPFlag.REPLY_EXPECTED, destination_port=0,
                destination_cpu=0,
                destination_chip_x=self.DEFAULT_DEST_X_COORD,
                destination_chip_y=self.DEFAULT_DEST_Y_COORD),
            SCPRequestHeader(command=SCPCommand.CMD_SIG),
            argument_1=_COUNT_SIGNAL_TYPE,
            argument_2=_get_data(app_id, state),
            argument_3=_ALL_CORE_MASK)

    def get_scp_response(self):
        return SCPCountStateResponse()



from six import add_metaclass

from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spinnman.messages.sdp.sdp_header import SDPHeader
from spinnman.messages.scp.scp_response_header import SCPResponseHeader

# The offset of the header from the start of a received packet
# (8 bytes of SDP header)
_SCP_HEADER_OFFSET = 8

# The offset of the data from the start of a received packet
# (8 bytes of SDP header + 4 bytes of SCP header)
_SCP_DATA_OFFSET = 12


@add_metaclass(AbstractBase)
class AbstractSCPResponse(object):
    """ Represents an abstract SCP Response
    """

    def __init__(self):
        """
        """
        self._sdp_header = None
        self._scp_response_header = None

    def read_bytestring(self, data, offset):
        """ Reads a packet from a bytestring of data

        :param data: The bytestring to be read
        :type data: str
        :param offset: The offset in the data from which the response should\
                    be read
        :type offset: int
        """
        self._sdp_header = SDPHeader.from_bytestring(data, offset)
        self._scp_response_header = SCPResponseHeader.from_bytestring(
            data, _SCP_HEADER_OFFSET + offset)
        self.read_data_bytestring(data, _SCP_DATA_OFFSET + offset)

    @abstractmethod
    def read_data_bytestring(self, data, offset):
        """ Reads the remainder of the data following the header

        :param data: The bytestring to read from
        :type data: str
        :param offset: The offset into the data after the headers
        :type offset: int
        """
        pass

    @property
    def sdp_header(self):
        """ The SDP header from the response

        :return: The SDP header
        :rtype: :py:class:`spinnman.messages.sdp.sdp_header.SDPHeader`
        :raise None: No known exceptions are raised
        """
        return self._sdp_header

    @property
    def scp_response_header(self):
        """ The SCP header from the response

        :return: The SCP header
        :rtype:\
                    :py:class:`spinnman.messages.scp.scp_response_header.SCPResponseHeader`
        :raise None: No known exceptions are raised
        """
        return self._scp_response_header



_MIN_APP_ID = 17
_MAX_APP_ID = 254


class AppIdTracker(object):
    """ A tracker of AppId to make it easier to allocate new ids
    """

    # Keep a class-global reference to the free id range, so ids are
    # allocated globally

    def __init__(
            self, app_ids_in_use=None, min_app_id=_MIN_APP_ID,
            max_app_id=_MAX_APP_ID):
        """

        :param app_ids_in_use: The ids that are already in use
        :type app_ids_in_use: list of int or None
        :param min_app_id: The smallest app id to use
        :type min_app_id: int
        :param max_app_id: The largest app id to use
        :type max_app_id: int
        """
        self._free_ids = set(range(min_app_id, max_app_id))
        if app_ids_in_use is not None:
            self._free_ids.difference_update(app_ids_in_use)
        self._min_app_id = min_app_id
        self._max_app_id = max_app_id

    def get_new_id(self):
        """ Get a new unallocated id

        :rtype: int
        """
        return self._free_ids.pop()

    def allocate_id(self, allocated_id):
        """ Allocate a given id - raises KeyError if the id is not present

        :param allocated_id: The id to allocate
        """
        self._free_ids.remove(allocated_id)

    def free_id(self, id_to_free):
        """ Free a given id - raises KeyError if the id is out of range

        :param id_to_free: The id to free
        """
        if id_to_free < self._min_app_id or id_to_free > self._max_app_id:
            raise KeyError("ID {} out of allowed range of {} to {}".format(
                id_to_free, self._min_app_id, self._max_app_id))
        self._free_ids.add(id_to_free)



from spinnman.connections.abstract_classes.abstract_spinnaker_boot_sender \
    import AbstractSpinnakerBootSender
from spinnman.connections.abstract_classes.abstract_spinnaker_boot_receiver \
    import AbstractSpinnakerBootReceiver
from spinnman.connections.udp_packet_connections.udp_connection import \
    UDPConnection
from spinnman.messages.spinnaker_boot.spinnaker_boot_message \
    import SpinnakerBootMessage
from spinnman import constants


class UDPBootConnection(UDPConnection,
                        AbstractSpinnakerBootSender,
                        AbstractSpinnakerBootReceiver):
    """ A connection to the spinnaker board that uses UDP to for booting
    """

    def __init__(self, local_host=None, local_port=None, remote_host=None,
                 remote_port=None):
        """
        :param local_host: The local host name or ip address to bind to.\
                    If not specified defaults to bind to all interfaces,\
                    unless remote_host is specified, in which case binding is\
                    _done to the ip address that will be used to send packets
        :type local_host: str
        :param local_port: The local port to bind to, between 1025 and 65535.\
                    If not specified, defaults to a random unused local port
        :type local_port: int
        :param remote_host: The remote host name or ip address to send packets\
                    to.  If not specified, the socket will be available for\
                    listening only, and will throw and exception if used for\
                    sending
        :type remote_host: str
        :param remote_port: The remote port to send packets to.  If\
                    remote_host is None, this is ignored.
        :raise spinnman.exceptions.SpinnmanIOException: If there is an error\
                    setting up the communication channel
        """

        if remote_port is None:
            remote_port = constants.UDP_BOOT_CONNECTION_DEFAULT_PORT

        UDPConnection.__init__(
            self, local_host, local_port, remote_host, remote_port)
        AbstractSpinnakerBootReceiver.__init__(self)
        AbstractSpinnakerBootSender.__init__(self)

    def send_boot_message(self, boot_message):
        """ See\
            :py:meth:`spinnman.connections.abstract_spinnaker_boot_sender.AbstractSpinnakerBootSender.send_boot_message`
        """
        self.send(boot_message.bytestring)

    def receive_boot_message(self, timeout=None):
        """ See\
            :py:meth:`spinnman.connections.abstract_spinnaker_boot_receiver.AbstractSpinnakerBootReceiver.receive_boot_message`
        """
        data = self.receive(timeout)
        return SpinnakerBootMessage.from_bytestring(data, 0)

    def __repr__(self):
        return\
            "UDPBootConnection(local_host={}, local_port={}, remote_host={},"\
            "remote_port={})".format(
                self.local_ip_address, self.local_port,
                self.remote_ip_address, self.remote_port)



from spinnman.messages.eieio.command_messages.eieio_command_message\
    import EIEIOCommandMessage
from spinnman.messages.eieio.command_messages.eieio_command_header\
    import EIEIOCommandHeader
from spinnman import constants


class StartRequests(EIEIOCommandMessage):
    """ Packet used in the context of buffering input for the host computer to\
        signal to the SpiNNaker system that, if needed, it is possible to\
        send more "SpinnakerRequestBuffers" packet
    """
    def __init__(self):
        EIEIOCommandMessage.__init__(
            self, EIEIOCommandHeader(
                constants.EIEIO_COMMAND_IDS.START_SENDING_REQUESTS.value))



import struct
from spinnman.messages.sdp.sdp_flag import SDPFlag

N_BYTES = 8


class SDPHeader(object):
    """ Represents the header of an SDP message.
        Each optional parameter in the constructor can be set to a value other\
        than None once, after which it is immutable.  It is an error to set a\
        parameter that is not currently None.
    """

    def __init__(self, flags=None, tag=None,
                 destination_port=None, destination_cpu=None,
                 destination_chip_x=None, destination_chip_y=None,
                 source_port=None, source_cpu=None,
                 source_chip_x=None, source_chip_y=None):
        """
        :param flags: Any flags for the packet
        :type flags: :py:class:`spinnman.messages.sdp.sdp_flag.SDPFlag`
        :param tag: The ip tag of the packet between 0 and 255, or None if it\
                    is to be set later
        :type tag: int
        :param destination_port: The destination port of the packet between 0\
                    and 7
        :type destination_port: int
        :param destination_cpu: The destination processor id within the chip\
                    between 0 and 31
        :type destination_cpu: int
        :param destination_chip_x: The x-coordinate of the destination chip\
                    between 0 and 255
        :type destination_chip_x: int
        :param destination_chip_y: The y-coordinate of the destination chip\
                    between 0 and 255
        :type destination_chip_y: int
        :param source_port: The source port of the packet between 0 and 7, or\
                    None if it is to be set later
        :type source_port: int
        :param source_cpu: The source processor id within the chip\
                   between 0 and 31, or None if it is to be set later
        :type source_cpu: int
        :param source_chip_x: The x-coordinate of the source chip\
                    between 0 and 255, or None if it is to be set later
        :type source_chip_x: int
        :param source_chip_y: The y-coordinate of the source chip\
                    between 0 and 255, or None if it is to be set later
        """
        self._flags = flags
        self._tag = tag
        self._destination_port = destination_port
        self._destination_cpu = destination_cpu
        self._destination_chip_x = destination_chip_x
        self._destination_chip_y = destination_chip_y
        self._source_port = source_port
        self._source_cpu = source_cpu
        self._source_chip_x = source_chip_x
        self._source_chip_y = source_chip_y

    @property
    def flags(self):
        """ The flags of the packet

        :return: The flags of the packet
        :rtype: :py:class:`spinnman.messages.sdp.sdp_flag.SDPFlag`
        """
        return self._flags

    @flags.setter
    def flags(self, flags):
        """ Set the flags of the packet

        :param flags: The flags to set
        :type flags: :py:class:`spinnman.messages.sdp.sdp_flag.SDPFlag`
        """
        self._flags = flags

    @property
    def tag(self):
        """ The tag of the packet

        :return: The tag of the packet between 0 and 255
        :rtype: int
        """
        return self._tag

    @tag.setter
    def tag(self, tag):
        """ Set the tag of the packet

        :param tag: The tag to set, between 0 and 255
        :type tag: int
        """
        self._tag = tag

    @property
    def destination_port(self):
        """ The destination port of the packet

        :return: The destination port of the packet between 0 and 7
        :rtype: int
        """
        return self._destination_port

    @destination_port.setter
    def destination_port(self, destination_port):
        """ Set the destination port of the packet

        :param destination_port: The destination port to set, between 0 and 7
        :type destination_port: int
        """
        self._destination_port = destination_port

    @property
    def destination_cpu(self):
        """ The core on the destination chip

        :return: The core on the destination chip, between 0 and 31
        :rtype: int
        """
        return self._destination_cpu

    @destination_cpu.setter
    def destination_cpu(self, destination_cpu):
        """ Set the id of the destination processor of the packet

        :param destination_cpu: The processor id to set, between 0 and 31
        :type destination_cpu: int
        """
        self._destination_cpu = destination_cpu

    @property
    def destination_chip_x(self):
        """ The x-coordinate of the destination chip of the packet

        :return: The x-coordinate of the chip, between 0 and 255
        :rtype: int
        """
        return self._destination_chip_x

    @destination_chip_x.setter
    def destination_chip_x(self, destination_chip_x):
        """ Set the x-coordinate of the destination chip of the packet

        :param destination_chip_x: The x-coordinate to set, between 0 and 255
        :type destination_chip_x: int
        """
        self._destination_chip_x = destination_chip_x

    @property
    def destination_chip_y(self):
        """ The y-coordinate of the destination chip of the packet

        :return: The y-coordinate of the chip, between 0 and 255
        :rtype: int
        """
        return self._destination_chip_y

    @destination_chip_y.setter
    def destination_chip_y(self, destination_chip_y):
        """ Set the y-coordinate of the destination chip of the packet

        :param destination_chip_y: The y-coordinate to set, between 0 and 255
        :type destination_chip_y: int
        """
        self._destination_chip_y = destination_chip_y

    @property
    def source_port(self):
        """ The source port of the packet

        :return: The source port of the packet between 0 and 7
        :rtype: int
        """
        return self._source_port

    @source_port.setter
    def source_port(self, source_port):
        """ Set the source port of the packet

        :param source_port: The source port to set, between 0 and 7
        :type source_port: int
        """
        self._source_port = source_port

    @property
    def source_cpu(self):
        """ The core on the source chip

        :return: The core on the source chip, between 0 and 31
        :rtype: int
        """
        return self._source_cpu

    @source_cpu.setter
    def source_cpu(self, source_cpu):
        """ Set the id of the source processor of the packet

        :param source_cpu: The processor id to set, between 0 and 31
        :type source_cpu: int
        """
        self._source_cpu = source_cpu

    @property
    def source_chip_x(self):
        """ The x-coordinate of the source chip of the packet

        :return: The x-coordinate of the chip, between 0 and 255
        :rtype: int
        """
        return self._source_chip_x

    @source_chip_x.setter
    def source_chip_x(self, source_chip_x):
        """ Set the x-coordinate of the source chip of the packet

        :param source_chip_x: The x-coordinate to set, between 0 and 255
        :type source_chip_x: int
        """
        self._source_chip_x = source_chip_x

    @property
    def source_chip_y(self):
        """ The y-coordinate of the source chip of the packet

        :return: The y-coordinate of the chip, between 0 and 255
        :rtype: int
        """
        return self._source_chip_y

    @source_chip_y.setter
    def source_chip_y(self, source_chip_y):
        """ Set the y-coordinate of the source chip of the packet

        :param source_chip_y: The y-coordinate to set, between 0 and 255
        :type source_chip_y: int
        """
        self._source_chip_y = source_chip_y

    @property
    def bytestring(self):
        """ The header as a bytestring

        :return: The header bytestring
        :rtype: str
        """
        dest_port_cpu = (((self._destination_port & 0x7) << 5) |
                         (self._destination_cpu & 0x1F))
        source_port_cpu = (((self._source_port & 0x7) << 5) |
                           (self._source_cpu & 0x1F))

        return struct.pack(
            "<8B", self._flags.value, self._tag, dest_port_cpu,
            source_port_cpu, self._destination_chip_y,
            self._destination_chip_x, self._source_chip_y, self._source_chip_x)

    @staticmethod
    def from_bytestring(data, offset):
        """ Read the header from a bytestring

        :param data: The bytestring to read the header from
        :type data: str
        :param offset: The offset into the data from which to start reading
        :type offset: int
        """

        (flags, tag, dest_port_cpu, source_port_cpu,
         destination_chip_y, destination_chip_x,
         source_chip_y, source_chip_x) = struct.unpack_from(
            "<8B", data, offset)
        destination_port = dest_port_cpu >> 5
        destination_cpu = dest_port_cpu & 0x1F
        source_port = source_port_cpu >> 5
        source_cpu = source_port_cpu & 0x1F
        return SDPHeader(
            SDPFlag(flags), tag, destination_port, destination_cpu,
            destination_chip_x, destination_chip_y, source_port, source_cpu,
            source_chip_x, source_chip_y)



from spinnman import constants


class SocketAddressWithChip(object):
    """ The address of a socket and an associated chip
    """

    def __init__(self, hostname, chip_x, chip_y,
                 port_num=constants.SCP_SCAMP_PORT):
        self._hostname = hostname
        self._port_num = port_num
        self._chip_x = chip_x
        self._chip_y = chip_y

    @property
    def hostname(self):
        """ The hostname of the socket

        :return: the hostname
        """
        return self._hostname

    @property
    def port_num(self):
        """ The port number of the socket

        :return: the port
        """
        return self._port_num

    @property
    def chip_x(self):
        """ The x-coordinate of the chip

        :return: the x-coordinate
        """
        return self._chip_x

    @property
    def chip_y(self):
        """ The y-coordinate of the chip

        :return: the y-coordinate
        """
        return self._chip_y

    def __str__(self):
        return "{}:{}:{}:{}".format(self._hostname, self._port_num,
                                    self._chip_x, self._chip_y)



import struct

from spinnman.messages.scp.enums.scp_result import SCPResult


class SCPResponseHeader(object):
    """ Represents the header of an SCP Response
    """

    def __init__(self, result=None, sequence=None):
        """
        """
        self._result = result
        self._sequence = sequence

    @property
    def result(self):
        """ The result of the SCP response

        :return: The result
        :rtype: :py:class:`spinnman.messages.scp.scp_result.SCPResult`
        """
        return self._result

    @property
    def sequence(self):
        """ The sequence number of the SCP response

        :return: The sequence number of the packet, between 0 and 65535
        :rtype: int
        """
        return self._sequence

    @staticmethod
    def from_bytestring(data, offset):
        """ Read a header from a bytestring

        :param data: The bytestring to read from
        :type data: str
        :param offset:
        """
        result, sequence = struct.unpack_from("<2H", data, offset)
        return SCPResponseHeader(SCPResult(result), sequence)



class BMPConnectionData(object):
    """ Contains the details of a BMP connection
    """

    def __init__(self, cabinet, frame, ip_address, boards, port_num):
        self._cabinet = cabinet
        self._frame = frame
        self._ip_address = ip_address
        self._boards = boards
        self._port_num = port_num

    @property
    def cabinet(self):
        """ Get the cabinet number

        :rtype: int
        """
        return self._cabinet

    @property
    def frame(self):
        """ Get the frame number

        :rtype: int
        """
        return self._frame

    @property
    def ip_address(self):
        """ Get the IP address of the BMP

        :rtype: str
        """
        return self._ip_address

    @property
    def boards(self):
        """ Iterable of integers for the boards to be addressed

        :rtype: iterable of int
        """
        return self._boards

    @property
    def port_num(self):
        """ The port number associated with this BMP connection

        :return: The port number
        """
        return self._port_num

    def __str__(self):
        return "{}:{}:{}:{}:{}".format(
            self._cabinet, self._frame, self._ip_address, self._boards,
            self._port_num)

    def __repr__(self):
        return self.__str__()



from spinnman.messages.scp.impl.scp_read_memory_request \
    import SCPReadMemoryRequest
from spinnman.model.router_diagnostics import RouterDiagnostics
from spinnman.processes.abstract_multi_connection_process \
    import AbstractMultiConnectionProcess

import struct

_N_REGISTERS = 16


class ReadRouterDiagnosticsProcess(AbstractMultiConnectionProcess):

    def __init__(self, connection_selector):
        AbstractMultiConnectionProcess.__init__(self, connection_selector)
        self._control_register = None
        self._error_status = None
        self._register_values = [0] * _N_REGISTERS

    def handle_control_register_response(self, response):
        self._control_register = struct.unpack_from(
            "<I", response.data, response.offset)[0]

    def handle_error_status_response(self, response):
        self._error_status = struct.unpack_from(
            "<I", response.data, response.offset)[0]

    def handle_register_response(self, response):
        for register in range(_N_REGISTERS):
            self._register_values[register] = struct.unpack_from(
                "<I", response.data, response.offset + (register * 4))[0]

    def get_router_diagnostics(self, x, y):
        self._send_request(SCPReadMemoryRequest(x, y, 0xe1000000, 4),
                           self.handle_control_register_response)
        self._send_request(SCPReadMemoryRequest(x, y, 0xe1000014, 4),
                           self.handle_error_status_response)
        self._send_request(SCPReadMemoryRequest(x, y, 0xe1000300, 16 * 4),
                           self.handle_register_response)
        self._finish()
        self.check_for_error()
        return RouterDiagnostics(self._control_register, self._error_status,
                                 self._register_values)



from spinnman.connections.udp_packet_connections.udp_connection \
    import UDPConnection
from spinnman.messages.sdp.sdp_message import SDPMessage
from spinnman.messages.sdp.sdp_flag import SDPFlag
from spinnman.connections.udp_packet_connections import udp_utils
from spinnman.connections.abstract_classes.abstract_sdp_receiver \
    import AbstractSDPReceiver
from spinnman.connections.abstract_classes.abstract_sdp_sender \
    import AbstractSDPSender
from spinnman.connections.abstract_classes.abstract_listenable \
    import AbstractListenable

import struct


class UDPSDPConnection(UDPConnection, AbstractSDPReceiver, AbstractSDPSender,
                       AbstractListenable):

    def __init__(self, chip_x=None, chip_y=None, local_host=None,
                 local_port=None, remote_host=None, remote_port=None):
        """
        :param chip_x: The optional x-coordinate of the chip at the remote\
                end of the connection.  If not specified, it will not be\
                possible to send SDP messages that require a response with\
                this connection.
        :type chip_x: int
        :param chip_y: The optional y-coordinate of the chip at the remote\
                end of the connection.  If not specified, it will not be\
                possible to send SDP messages that require a response with\
                this connection.
        :type chip_y: int
        :param local_host: The optional ip address or host name of the local\
                interface to listen on
        :type local_host: str
        :param local_port: The optional local port to listen on
        :type local_port: int
        :param remote_host: The optional remote host name or ip address to\
                send messages to.  If not specified, sending will not be\
                possible using this connection
        :type remote_host: str
        :param remote_port: The optional remote port number to send messages\
                to.  If not specified, sending will not be possible using this\
                connection
        """
        UDPConnection.__init__(
            self, local_host, local_port, remote_host, remote_port)
        AbstractSDPReceiver.__init__(self)
        AbstractSDPSender.__init__(self)
        AbstractListenable.__init__(self)
        self._chip_x = chip_x
        self._chip_y = chip_y

    def receive_sdp_message(self, timeout=None):
        data = self.receive(timeout)
        return SDPMessage.from_bytestring(data, 2)

    def send_sdp_message(self, sdp_message):

        # If a reply is expected, the connection should
        if sdp_message.sdp_header.flags == SDPFlag.REPLY_EXPECTED:
            udp_utils.update_sdp_header_for_udp_send(
                sdp_message.sdp_header, self._chip_x, self._chip_y)
        else:
            udp_utils.update_sdp_header_for_udp_send(
                sdp_message.sdp_header, 0, 0)
        self.send(struct.pack("<2x") + sdp_message.bytestring)

    def get_receive_method(self):
        return self.receive_sdp_message

    def __repr__(self):
        return \
            "UDPSDPConnection(chip_x={}, chip_y={}, local_host={},"\
            " local_port={}, remote_host={}, remote_port={})".format(
                self._chip_x, self._chip_y, self.local_ip_address,
                self.local_port, self.remote_ip_address, self.remote_port)



from spinnman.messages.scp.abstract_messages.abstract_scp_bmp_reponse import \
    AbstractSCPBMPResponse
from spinnman.messages.scp.enums.scp_result import SCPResult
from spinnman.exceptions import SpinnmanUnexpectedResponseCodeException
from spinnman.model.adc_info import ADCInfo


class SCPReadADCResponse(AbstractSCPBMPResponse):
    """ An SCP response to a request for ADC information
    """

    def __init__(self):
        """
        """
        AbstractSCPBMPResponse.__init__(self)
        self._adc_info = None

    def read_data_bytestring(self, data, offset):
        """ See\
            :py:meth:`spinnman.messages.scp.abstract_scp_response.AbstractSCPResponse.read_data_bytestring`
        """
        result = self.scp_response_header.result
        if result != SCPResult.RC_OK:
            raise SpinnmanUnexpectedResponseCodeException(
                "ADC", "CMD_ADC", result.name)
        self._adc_info = ADCInfo(data, offset)

    @property
    def adc_info(self):
        """ The ADC information
        """
        return self._adc_info



from spinnman.messages.spinnaker_boot._system_variables.\
    _system_variable_boot_values import SystemVariableDefinition

import struct

# The base address of the system variable structure in System ram
_SYSTEM_VARIABLE_BASE_ADDRESS = 0xf5007f00

# The size of the system variable structure in bytes
_SYSTEM_VARIABLE_BYTES = 256


class ChipInfo(object):
    """ Represents the system variables for a chip, received from the chip\
        SDRAM
    """

    def __init__(self, system_data, offset):
        """

        :param system_data: An bytestring retrieved from SDRAM on the board
        :type system_data: str
        :param offset: The offset into the bytestring where the actual data\
                starts
        :raise spinnman.exceptions.SpinnmanInvalidParameterException: If\
                    the data doesn't contain valid system data information
        """
        self._system_data = system_data
        self._offset = offset

        links_available = self._read_value("links_available")
        self._links_available = list()
        for i in range(0, 6):
            if ((links_available >> i) & 0x1) != 0:
                self._links_available.append(i)

        self._led_flash_period_ms = self._read_value(
            "led_half_period_10_ms") * 10
        self._leds = [self._read_value("led_0"), self._read_value("led_1")]
        self._status_map = bytearray(self._read_value("status_map"))
        self._physical_to_virtual_core_map = bytearray(
            self._read_value("physical_to_virtual_core_map"))
        self._virtual_to_physical_core_map = bytearray(
            self._read_value("virtual_to_physical_core_map"))

        self._virtual_core_ids = list()
        for physical_core_id in range(
                0, len(self._physical_to_virtual_core_map)):
            virtual_core_id = self._physical_to_virtual_core_map[
                physical_core_id]
            if virtual_core_id != 0xFF:
                self._virtual_core_ids.append(virtual_core_id)
        self._virtual_core_ids.sort()

        ip = bytearray(self._read_value("ethernet_ip_address"))
        self._ip_address = "{}.{}.{}.{}".format(ip[0], ip[1], ip[2], ip[3])
        if self._ip_address == "0.0.0.0":
            self._ip_address = None

    def _read_value(self, item):
        item_def = SystemVariableDefinition[item]
        code = item_def.data_type.struct_code
        if item_def.array_size is not None:
            code = "{}{}".format(item_def.array_size, code)
        values = struct.unpack_from(
            code, self._system_data, self._offset + item_def.offset)
        return values[0]

    def __getattr__(self, item):
        return self._read_value(item)

    @property
    def x(self):
        """ The x-coordinate of the chip

        :rtype: int
        """
        return self._read_value("x")

    @property
    def y(self):
        """ The y-coordinate of the chip

        :rtype: int
        """
        return self._read_value("y")

    @property
    def x_size(self):
        """ The number of chips in the x-dimension

        :rtype: int
        """
        return self._read_value("x_size")

    @property
    def y_size(self):
        """ The number of chips in the y-dimension

        :rtype: int
        """
        return self._read_value("y_size")

    @property
    def nearest_ethernet_x(self):
        """ The x-coordinate of the nearest chip with Ethernet

        :rtype: int
        """
        return self._read_value("nearest_ethernet_x")

    @property
    def nearest_ethernet_y(self):
        """ The y-coordinate of the nearest chip with Ethernet

        :rtype: int
        """
        return self._read_value("nearest_ethernet_y")

    @property
    def is_ethernet_available(self):
        """ True if the Ethernet is running on this chip, False otherwise

        :rtype: bool
        """
        return self._read_value("is_ethernet_available") == 1

    @property
    def links_available(self):
        """ The links that are available on the chip

        :rtype: iterable of int
        """
        return self._links_available

    @property
    def cpu_clock_mhz(self):
        """ The speed of the CPU clock in MHz

        :rtype: int
        """
        return self._read_value("cpu_clock_mhz")

    @property
    def physical_to_virtual_core_map(self):
        """ The physical core id to virtual core id map; entries with a value\
            of 0xFF are non-operational cores

        :rtype: bytearray
        """
        return self._physical_to_virtual_core_map

    @property
    def virtual_core_ids(self):
        """ A list of available cores by virtual core id (including the\
            monitor)

        :rtype: iterable of int
        """
        return self._virtual_core_ids

    @property
    def sdram_base_address(self):
        """ The base address of the user region of SDRAM on the chip

        :rtype: int
        """
        return self._read_value("sdram_base_address")

    @property
    def system_sdram_base_address(self):
        """ The base address of the System SDRAM region on the chip

        :rtype: int
        """
        return self._read_value("system_sdram_base_address")

    @property
    def cpu_information_base_address(self):
        """ The base address of the cpu information structure

        :rtype: int
        """
        return self._read_value("cpu_information_base_address")

    @property
    def first_free_router_entry(self):
        """ The id of the first free routing entry on the chip

        :rtype: int
        """
        return self._read_value("first_free_router_entry")

    @property
    def ip_address(self):
        """ The ip address of the chip, or None if no Ethernet

        :rtype: str
        """
        return self._ip_address

    @property
    def iobuf_size(self):
        """ The size of the iobuf buffers in bytes

        :rtype: int
        """
        return self._read_value("iobuf_size")

    def router_table_copy_address(self):
        """ The address of the copy of the router table

        :rtype: int
        """
        return self._read_value("router_table_copy_address")

    @property
    def system_ram_heap_address(self):
        """ The address of the base of the heap in system RAM

        :rtype: int
        """
        return self._read_value("system_ram_heap_address")

    @property
    def sdram_heap_address(self):
        """ The address of the base of the heap in SDRAM

        :rtype: int
        """
        return self._read_value("sdram_heap_address")



from spinnman import constants

from spinnman.messages.scp.abstract_messages.abstract_scp_request\
    import AbstractSCPRequest
from spinnman.messages.sdp.sdp_flag import SDPFlag
from spinnman.messages.sdp.sdp_header import SDPHeader
from spinnman.messages.scp.scp_request_header import SCPRequestHeader
from spinnman.messages.scp.enums.scp_command import SCPCommand
from spinnman.messages.scp.impl.scp_read_memory_response import \
    SCPReadMemoryResponse


class SCPReadMemoryRequest(AbstractSCPRequest):
    """ An SCP request to read a region of memory on a chip
    """

    def __init__(self, x, y, base_address, size, cpu=0):
        """

        :param x: The x-coordinate of the chip to read from, between 0 and 255
        :type x: int
        :param y: The y-coordinate of the chip to read from, between 0 and 255
        :type y: int
        :param base_address: The positive base address to start the read from
        :type base_address: int
        :param size: The number of bytes to read, between 1 and 256
        :type size: int
        :raise spinnman.exceptions.SpinnmanInvalidParameterException:
                    * If the chip coordinates are out of range
                    * If the base address is not a positive number
                    * If the size is out of range
        """
        super(SCPReadMemoryRequest, self).__init__(
            SDPHeader(
                flags=SDPFlag.REPLY_EXPECTED, destination_port=0,
                destination_cpu=cpu, destination_chip_x=x,
                destination_chip_y=y),
            SCPRequestHeader(command=SCPCommand.CMD_READ),
            argument_1=base_address, argument_2=size,
            argument_3=constants.address_length_dtype[
                (base_address % 4, size % 4)].value)

    def get_scp_response(self):
        """ See\
            :py:meth:`spinnman.messages.scp.abstract_scp_request.AbstractSCPRequest.get_scp_response`
        """
        return SCPReadMemoryResponse()



from enum import Enum


class SDPFlag(Enum):
    """ SDPFlag for the message
    """
    REPLY_NOT_EXPECTED = (0x07, "Indicates that a reply is not expected")
    REPLY_EXPECTED = (0x87, "Indicates that a reply is expected")

    def __new__(cls, value, doc=""):
        obj = object.__new__(cls)
        obj._value_ = value
        return obj

    def __init__(self, value, doc=""):
        self._value_ = value
        self.__doc__ = doc



from six import add_metaclass

from spinn_utilities.abstract_base import AbstractBase, abstractmethod


@add_metaclass(AbstractBase)
class AbstractConnection(object):
    """ An abstract connection to the SpiNNaker board over some medium
    """

    __slots__ = ()

    @abstractmethod
    def is_connected(self):
        """ Determines if the medium is connected at this point in time

        :return: True if the medium is connected, False otherwise
        :rtype: bool
        :raise spinnman.exceptions.SpinnmanIOException: If there is an error\
                    when determining the connectivity of the medium
        """
        pass

    @abstractmethod
    def close(self):
        """ Closes the connection

        :return: Nothing is returned
        :rtype: None
        :raise None: No known exceptions are raised
        """
        pass



from spinnman.messages.scp.abstract_messages.abstract_scp_request\
    import AbstractSCPRequest
from spinnman.messages.sdp.sdp_header import SDPHeader
from spinnman.messages.sdp.sdp_flag import SDPFlag
from spinnman.messages.scp.scp_request_header import SCPRequestHeader
from spinnman.messages.scp.enums.scp_command import SCPCommand
from spinnman.messages.scp.impl.scp_check_ok_response import SCPCheckOKResponse


class SCPRouterClearRequest(AbstractSCPRequest):
    """ A request to clear the router on a chip
    """

    def __init__(self, x, y):
        """

        :param x: The x-coordinate of the chip, between 0 and 255 \
        this is not checked due to speed restrictions
        :type x: int
        :param y: The y-coordinate of the chip, between 0 and 255 \
        this is not checked due to speed restrictions
        :type y: int
        :raise spinnman.exceptions.SpinnmanInvalidParameterException:\
                    * If x is out of range
                    * If y is out of range
        """
        super(SCPRouterClearRequest, self).__init__(
            SDPHeader(
                flags=SDPFlag.REPLY_EXPECTED, destination_port=0,
                destination_cpu=0, destination_chip_x=x,
                destination_chip_y=y),
            SCPRequestHeader(command=SCPCommand.CMD_RTR))

    def get_scp_response(self):
        return SCPCheckOKResponse("RouterClear", "CMD_RTR")



from spinnman.messages.sdp.sdp_header import SDPHeader


class SDPMessage(object):
    """ Wraps up an SDP message with a header and optional data.
    """

    def __init__(self, sdp_header, data=None, offset=0):
        """
        :param sdp_header: The header of the message
        :type sdp_header:\
                    :py:class:`spinnman.messages.sdp.sdp_header.SDPHeader`
        :param data: The data of the SDP packet, or None if no data
        :type data: str
        :param offset: The offset where the valid data starts
        :type offset: int
        :raise None: No known exceptions are thrown
        """

        self._sdp_header = sdp_header
        self._data = data
        self._offset = offset

    @property
    def bytestring(self):
        """ The bytestring of the message

        :return: The bytestring of the message
        :rtype: str
        """
        if self._data is not None:
            return self._sdp_header.bytestring + self._data[self._offset:]
        return self._sdp_header.bytestring

    @staticmethod
    def from_bytestring(data, offset):
        sdp_header = SDPHeader.from_bytestring(data, offset)
        return SDPMessage(sdp_header, data, offset + 8)

    @property
    def sdp_header(self):
        """ The header of the packet

        :return: An SDP header
        :rtype: :py:class:`spinnman.messages.sdp.sdp_header.SDPHeader`
        """
        return self._sdp_header

    @property
    def data(self):
        """ The data in the packet

        :rtype: str
        """
        return self._data

    @property
    def offset(self):
        """ The offset where the valid data starts

        :rtype: int
        """
        return self._offset
